/*
 *	ghostscript.c
 *
 * These codes are roughly based on gs-view.c from miyu-patch for xdvi
 * PL17, psgs.c from xdvi PL20, and Ghostview.c from ghostview 1.5.
 *
 */

#include "xdvi.h"

#include <X11/Xatom.h>
#include <sys/stat.h>
#include <signal.h>
#include <X11/Intrinsic.h>

#undef	HAVE_SETENV
#define	USE_DUP2
#define	USE_FCLOSE

#ifndef	SIGCHLD
#define	SIGCHLD	SIGCLD
#endif

#ifdef	SIGNALRETURNSINT
#define	SIGVAL	int
#else
#define	SIGVAL	void
#endif

extern	int	errno;

#ifndef	SEEK_SET
#define	SEEK_SET	0
#endif

extern Pixel	fore_Pixel, back_Pixel;

static int  gs_initialized = 0;
static Atom	gs_atom;
static Atom gs_colors_atom;
static Atom	gs_done, gs_page;
static char **gs_argv = NULL;
static int gs_args;

static	jmp_buf	abort_gs;

/*
 * setenv() is stolen from xdvi PL20. Followings are original Copyright.
 */
/*
 * Copyright (c) 1994 Paul Vojta.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 *	Our replacement for setenv(), which is not available on all systems.
 */

#ifndef	HAVE_SETENV	/* define this if you're a performance freak and
			   if your system has setenv. */
#define	setenv(var, str, repl)	_setenv(var, str)	/* repl always True */

extern	char	**environ;

static	void
_setenv(var, str)
	_Xconst	char	*var;
	_Xconst	char	*str;
{
	int		len1;
	int		len2;
	char		*newvar;
	char		**linep;
	static	Boolean	malloced = False;

	len1 = strlen(var);
	len2 = strlen(str) + 1;
	newvar = xmalloc((unsigned int) len1 + len2 + 1, "_setenv");
	(void) bcopy(var, newvar, len1);
	newvar[len1++] = '=';
	(void) bcopy(str, newvar + len1, len2);
	for (linep = environ; *linep != NULL; ++linep)
	    if (memcmp(*linep, newvar, len1) == 0) {
		*linep = newvar;
		return;
	    }
	len1 = linep - environ;
	if (malloced) {
	    environ = (char **) realloc((char *) environ,
		(unsigned int) (len1 + 2) * sizeof(char *));
	    if (environ == NULL)
		oops("! Cannot allocate %d bytes for string list in _setenv.\n",
		    (len1 + 2) * sizeof(char *));
	}
	else {
	    linep = (char **) xmalloc((unsigned int)(len1 + 2) * sizeof(char *),
		"string list in _setenv");
	    (void) bcopy((char *) environ, (char *) linep,
		len1 * sizeof(char *));
	    environ = linep;
	    malloced = True;
	}
	environ[len1] = newvar;
	environ[len1 + 1] = NULL;
}
#endif	/* HAVE_SETENV */

int
round(x)
	float x;
{
	if (x > 0) return (int) (x + 0.5);
	else return (int) (x - 0.5);
}

static	SIGVAL
catch_sig_child()
{
	/* longjmp(abort_gs, SIGCHLD); */
#ifdef	SIGNALRETURNSINT
	return 0;
#endif
}

static	SIGVAL
catch_sig_user1()
{
	longjmp(abort_gs, SIGUSR1);
	/*NOTREACHED*/
}

#define	PSFIG_CACHE_SIZE	32

static struct PsfigCache {
	struct PsfigCache *next;
	Pixmap		pix;
	char		*file;
	int			width, height;
	time_t		mtime;
} *psfig_cache[PSFIG_CACHE_SIZE];

static Pixmap
get_cached_psfigure(file, width, height)
	char *file;
	int width, height;
{
	int 	hash = 0;
	int 	i;
	struct PsfigCache *p, *q;
	struct stat buf;

	for (i = 0; file[i]; i++)
	    hash += file[i];
	hash += width + height;
	hash %= PSFIG_CACHE_SIZE;


	if (psfig_cache[hash] == NULL) return (Pixmap)0;

	stat(file, &buf);

	for (p = psfig_cache[hash], q = NULL; p != NULL; q = p, p = p->next)
		if (p->width  == width  && p->height == height &&
			strcmp(p->file, file) == 0) {
			if (p->mtime == buf.st_mtime) {
				if (q) {
					q->next = p->next;
					p->next = psfig_cache[hash];
					psfig_cache[hash] = p;
				}
				return p->pix;
			}
			else {
				if (q) q->next = p->next;
				else psfig_cache[hash] = NULL;
				XFreePixmap(DISP, p->pix);
				free(p);
				return (Pixmap)0;
			}
		}

	return (Pixmap)0;
}


static	void
put_cache_psfigure(pix, file, width, height)
	Pixmap pix;
	char *file;
	int width, height;
{
	int 	hash = 0;
	int 	i;
	struct PsfigCache *p;
	struct stat buf;

	for (i = 0; file[i]; i++)
	    hash += file[i];
	hash += width + height;
	hash %= PSFIG_CACHE_SIZE;

	stat(file, &buf);

	p = (struct PsfigCache *)
		xmalloc(sizeof(struct PsfigCache), "caching PS");
	if (p == NULL) return;

	p->next = psfig_cache[hash];
	psfig_cache[hash] = p;
	p->pix = pix;
	p->file = (char *)xmalloc(strlen(file) + 1, "caching PS");
	Strcpy(p->file, file);
	p->width = width;
	p->height = height;
	p->mtime = buf.st_mtime;
}


static void
init_gs()
{
	int		i;
#ifdef	PS_HEADER
	struct ps_hdr	*p;
#endif	/* PS_HEADER */

	gs_args = 7;
#ifdef	PS_HEADER
	for (p = ps_headers; p != NULL; p = p->next) gs_args++;
#endif	/* PS_HEADER */
	if (gs_safer) gs_args++;
	gs_argv = (char **)xmalloc(sizeof(char *) * gs_args, "gs");

	i = 0;
	gs_argv[i++] = gs_path;
	gs_argv[i++] = "-dQUIET";
	gs_argv[i++] = "-dNOPAUSE";
	if (gs_safer) gs_argv[i++] = "-dSAFER";
#ifdef	PS_HEADER
	for (p = ps_headers; p != NULL; p = p->next)
		gs_argv[i++] = p->psheader;
#endif	/* PS_HEADER */
	gs_argv[i++] = "-";
	gs_argv[i++] = NULL;
	gs_argv[i++] = "quit.ps";
	gs_argv[i++] = NULL;

	gs_initialized = 1;
}
#define	GS_FILENAME		gs_argv[gs_args-3]

Bool
isgsevent(d, ev)
	Display	*d;
	XEvent	*ev;
{
	if (ev->type == ClientMessage &&
		ev->xclient.window == WINDOW(currwin) &&
		ev->xclient.send_event == 1) {
		if (ev->xclient.message_type == gs_done ||
			ev->xclient.message_type == gs_page)
			return True;
	}
	return False;
}

Pixmap
get_psfigure(file, width, height, llx, lly, urx, ury)
	char	*file;
	int	width, height;
	double	llx, lly, urx, ury;
{
	char	buf[BUFSIZ];
	int		gs_pid;
	Pixmap	pix;
	XEvent	ev;
	SIGVAL	(*old_sigchld)(), (*old_sigusr1)();

	if (pix = get_cached_psfigure(file, width, height))
		return pix;

	if (!gs_initialized) init_gs();

	gs_done = XInternAtom(DISP, "DONE", False);
	gs_page = XInternAtom(DISP, "PAGE", False);

	gs_atom = XInternAtom(DISP, "GHOSTVIEW", False);
	Sprintf(buf, "0 0 %d %d %d %d %g %g 0 0 0 0",
		round(llx), round(lly),
		round(urx), round(ury),
		width  / ((urx - llx) / 72.0),
		height / ((ury - lly) / 72.0));
	XChangeProperty(DISP, WINDOW(currwin), gs_atom, XA_STRING,
					8, PropModeReplace, buf, strlen(buf));
	gs_colors_atom = XInternAtom(DISP, "GHOSTVIEW_COLORS", False);
	if (*gs_palette == 'M')
		Sprintf(buf, "%s %ld %ld", gs_palette, fore_Pixel, back_Pixel);
	else
		Sprintf(buf, "%s 0 1", gs_palette);
	XChangeProperty(DISP, WINDOW(currwin), gs_colors_atom, XA_STRING,
					8, PropModeReplace, buf, strlen(buf));
	XSync(DISP, False);

	pix = XCreatePixmap(DISP, WINDOW(currwin),
						width, height, DefaultDepthOfScreen(SCRN));

#ifdef	PS_LITERAL
	fflush(temp_ps_file);
	lseek(fileno(temp_ps_file), 0L, SEEK_SET);
#endif	/* PS_LITERAL */
	{
		int val;

		val = setjmp(abort_gs);
		if (val == SIGUSR1) {
			Fprintf(stderr, "GS terminated abnormally.\n");
			wait(NULL);
#ifdef	PS_LITERAL
			temp_ps_flush();
#endif	/* PS_LITERAL */
			if (pix) XFreePixmap(DISP, pix);
			return (Pixmap)0;
		}
	}

	old_sigchld = signal(SIGCHLD, catch_sig_child);
	old_sigusr1 = signal(SIGUSR1, catch_sig_user1);

	if ((gs_pid = fork()) == 0) {
		struct font	*fontp;

#ifdef USE_DUP2
#ifdef PS_LITERAL
		if (dup2(fileno(temp_ps_file), fileno(stdin)) != fileno(stdin))
#else
		if (dup2(open("/dev/null", O_RDONLY), fileno(stdin)) != fileno(stdin))
#endif
			goto gs_exec_error;

		if (dup2(open("/dev/null", O_WRONLY), fileno(stdout))
			!= fileno(stdout))
			goto gs_exec_error;

#else /* USE_DUP2 */

#ifdef USE_FCLOSE
		fclose(stdin);
#else
		close(fileno(stdin));
#endif

#ifdef	PS_LITERAL
		if (dup(fileno(temp_ps_file)) != fileno(stdin))
#else
		if (open("/dev/null", O_RDONLY) != fileno(stdin))
#endif
			goto gs_exec_error;

#ifdef USE_FCLOSE
		fclose(stdout);
#else
		close(fileno(stdout));
#endif

		if (open("/dev/null", O_WRONLY) != fileno(stdout))
			goto gs_exec_error;
#endif /* USE_DUP2 */

	    for (fontp = font_head; fontp != NULL; fontp = fontp->next)
		if (fontp->file != NULL) {
		    Fclose(fontp->file);
		    fontp->file = NULL;
		    ++n_files_left;
		}

		Sprintf(buf, "%d %d", WINDOW(currwin), pix);
		setenv("GHOSTVIEW", buf, True);
		setenv("DISPLAY", XDisplayString(DISP), True);

		GS_FILENAME = file;
		execvp(gs_path, gs_argv);
		perror(gs_path);

gs_exec_error:
	    kill(getppid(), SIGUSR1);
	    Fprintf(stderr, "gs:exec error!\n");
	    exit(1);
	}

	XIfEvent(DISP, &ev, isgsevent, NULL);
	signal(SIGCHLD, old_sigchld);
	signal(SIGUSR1, old_sigusr1);
	kill(gs_pid, SIGTERM);
	wait(NULL);

#ifdef	PS_LITERAL
	temp_ps_flush();
#endif	/* PS_LITERAL */
	put_cache_psfigure(pix, file, width, height);

	return pix;
}

